/*
 * This file is part of PAXLink, which is an AI for OpenTTD
 * Copyright (C) 2009  Leif Linse
 *
 * PAXLink is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License
 *
 * PAXLink is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with PAXLink; If not, see <http://www.gnu.org/licenses/> or
 * write to the Free Software Foundation, Inc., 51 Franklin Street, 
 * Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
require("helper.nut");

class OrderList {
	list = null;

	skip_to_last_when_full = false;
	
	constructor()
	{
		this.list = [];
		skip_to_last_when_full = false;
	}
}

// AddStop function is at the very bottom, due to the table syntax is not really compatible with my hacked VIM systax file

function OrderList::SkipToLastWhenFull(enable)
{
	skip_to_last_when_full = enable;
}

function OrderList::FindOrderDestination(vehicle_id, begin, match_destination)
{
	local num = AIOrder.GetOrderCount(vehicle_id);
	for(local i = begin; i < num; i++)
	{
		if(AIOrder.GetOrderDestination(vehicle_id, i) == match_destination)
			return i;
	}

	return -1;
}

function OrderList::ApplyToVehicle(vehicle_id)
{
	AILog.Info("order list len: " + this.list.len());
	//Helper.BreakPoint(AIVehicle.GetLocation(vehicle_id));

	local i = 0;
	for(; i < list.len(); i++)
	{
		AILog.Info("i = " + i);
		if(i < AIOrder.GetOrderCount(vehicle_id))
		{
			AILog.Info(" i < num orders of vehicle");

			// If current order is a conditional order
			if(AIOrder.IsConditionalOrder(vehicle_id, i))
			{
				// remove conditional order
				AIOrder.RemoveOrder(vehicle_id, i);

				// skip to next order. Since one order was removed next order now has same id as the removed order (if it exists)
				i--;
				continue;

				// then at the end we add back conditional orders
			}

			local curr_order_dest_station = AIStation.GetStationID(AIOrder.GetOrderDestination(vehicle_id, i));
			if(curr_order_dest_station == this.list[i].station_id)
			{
				// current order has correct destination
				AILog.Info("current order has correct destination");
				AIOrder.ChangeOrder(vehicle_id, i, this.list[i].flags); // Make sure the flags are correct
			}
			else
			{
				// current order has wrong destination
				AILog.Info("current order has Wrong destination");

				// First try to see if there is an order with this destination
				local later_order = FindOrderDestination(vehicle_id, i+1, this.list[i].station_id);
				if(later_order != -1)
				{
					// Found the destination later in list, move it to i
					AILog.Info("Found correct destination later in list");
					AIOrder.MoveOrder(vehicle_id, later_order, i);
					AIOrder.ChangeOrder(vehicle_id, i, this.list[i].flags); // Make sure the flags are correct
				}
				else
				{
					// There is no order with this destination, make a new order.
					AILog.Info("Did not found correct destination later in list -> insert a new one");
					Helper.SetSign(AIStation.GetLocation(this.list[i].station_id), "insert");
					AIOrder.InsertOrder(vehicle_id, i, AIStation.GetLocation(this.list[i].station_id), this.list[i].flags);
					AILog.Info("Insert order error msg: " + AIError.GetLastErrorString());
				}
			}
		}
		else
		{
			// the existing order list is shorter than current wanted order.
			AILog.Info("Existing order list is shorter than current i -> Append new order");
			
			// Append wanted order at the end
			Helper.SetSign(AIStation.GetLocation(this.list[i].station_id), "append");
			AIOrder.AppendOrder(vehicle_id, AIStation.GetLocation(this.list[i].station_id), this.list[i].flags);
			AILog.Info("Append order error msg: " + AIError.GetLastErrorString());
		}

		AILog.Info("Breaking before next loop");
		//Helper.BreakPoint(AIVehicle.GetLocation(vehicle_id));
	}

	AILog.Info("Breaking before removal");
	//Helper.BreakPoint(AIVehicle.GetLocation(vehicle_id));

	// if the vehicle order list contained orders that we don't want to have anymore, they will be located at the end.
	// so remove all orders that are after the last item number in this.list
	local num_to_remove = AIOrder.GetOrderCount(vehicle_id) - this.list.len();
	if(num_to_remove > 0)
	{
		for(local j = 0; j < num_to_remove; j++)
		{
			// Remove the one with id last wanted + 1, since the remaining ones move up after each removal.
			AIOrder.RemoveOrder(vehicle_id, this.list.len());
		}
	}

	AILog.Info("Breaking after removal");
	//Helper.BreakPoint(AIVehicle.GetLocation(vehicle_id));

	// Add conditional orders if wanted
	if(skip_to_last_when_full && AIOrder.GetOrderCount(vehicle_id) >= 2)
	{
		// for all but last
		local order = 0;
		local num_regular_orders = AIOrder.GetOrderCount(vehicle_id);
		AILog.Info("Num regular orders: " + num_regular_orders);
		for(local i = 0; i < num_regular_orders - 1; i++)
		{
			AILog.Info("Inserting a conditional order at location " + (order + 1));
			AIOrder.InsertConditionalOrder(vehicle_id, order + 1, AIOrder.GetOrderCount(vehicle_id) - 1);
			AIOrder.SetOrderCondition(vehicle_id, order + 1, AIOrder.OC_LOAD_PERCENTAGE);
			AIOrder.SetOrderCompareFunction(vehicle_id, order + 1, AIOrder.CF_EQUALS);
			AIOrder.SetOrderCompareValue(vehicle_id, order + 1, 100);

			AILog.Info("Order count: " + AIOrder.GetOrderCount(vehicle_id));

			order += 2; // next regular order

			AILog.Info("i = " + i);
		}
	}
}

function OrderList::AddStop(station_id_value, flags_value)
{
	list.append(
		{station_id=station_id_value, 
		flags=flags_value}
	);
}

